/*
 * neural.cpp
 *
 *  Created on: May 26, 2009
 *      Author: vfedyashov
 */
#include "neural.h"


double G(double y)
{
	return (exp(y) - exp(-y))/(exp(y) + exp(-y));
}

double difG(double x)
{
	return 1 - x * x;
}


void NeuralNetwork::Init(int number_Of_Layers, int *neurons_Per_Layer)
{

}

void Layer::Calculate()
{
    VectorNeurons::iterator n_iter;
    VectorConnections::iterator c_iter;

    double tmp_Sum;

    for( n_iter=m_Neurons.begin(); n_iter<m_Neurons.end(); n_iter++ )
    {
        Neuron& n = *(*n_iter);


        c_iter = n.m_Connections.begin();

        tmp_Sum = m_Weights[(*c_iter)->WeightIndex]->value;

        for ( c_iter++ ; c_iter<n.m_Connections.end(); c_iter++ )
        {

            tmp_Sum += ( m_Weights[(*c_iter)->WeightIndex]->value ) *
                ( m_pPrevLayer->m_Neurons[ (*c_iter)->NeuronIndex ]->output );
        }

         n.output = G(tmp_Sum );

    }
}

void NeuralNetwork::FeedForward(double* inputVector, int iCount,
               double* outputVector,
               int oCount)

{
    VectorLayers::iterator l_iter = m_Layers.begin();
    VectorNeurons::iterator n_iter;

    //
    if ( l_iter < m_Layers.end() )
    {
        n_iter = (*l_iter)->m_Neurons.begin();
        int count = 0;


        //iCount = amount of neurons in the input layer, always
        while( ( n_iter < (*l_iter)->m_Neurons.end() ) && ( count < iCount ) )
        {
            (*n_iter)->output = inputVector[count];
            n_iter++;
            count++;
        }
    }

    // iterate through remaining layers

    for( l_iter++; l_iter < m_Layers.end(); l_iter++ )
    {
        (*l_iter)->Calculate();
    }

    // Results

    if (outputVector != NULL )
    {
        l_iter = m_Layers.end();
        l_iter--;

        n_iter = (*l_iter)->m_Neurons.begin();

        for ( int ii=0; ii < oCount; ++ii )
        {
            outputVector[ ii ] = (*n_iter)->output;
            n_iter++;
        }
    }
}

// Layer BackPropagnition

void Layer::Backpropagate(vector< double >& ddXn /* in */,
                            vector< double >& ddXprev /* out */,
                            double etaLearningRate )
{
    double output;

    // calculate dE/dYn;

    vector< double > ddYn(ddXprev.size());

    for (int ii = 0; ii < m_Neurons.size(); ++ii)
    {
        output = m_Neurons[ii]->output;
        ddYn[ii ] = difG(output) * ddXn[ii];
    }

    // Block below is to calculate dE/dWn for every i in (0..m_Weights-1)

    VectorNeurons::iterator n_iter;
    VectorConnections::iterator c_iter;
    int ii = 0;
    vector< double > ddWn(m_Weights.size());
    for (n_iter = m_Neurons.begin(); n_iter < m_Neurons.end(); n_iter++ )
    {
        Neuron& n = *(*n_iter);  // for simplifying the terminology


        for (c_iter = n.m_Connections.begin(); c_iter < n.m_Connections.end(); c_iter++ )
        {
            int kk = (*c_iter)->NeuronIndex;
            if ( kk == ULONG_MAX )
            {
                output = 1.0;  // this is the bias weight
            }
            else
            {
                output = m_pPrevLayer->m_Neurons[ kk ]->output;
            }

            ddWn[(*c_iter)->WeightIndex ] += ddYn[ ii ] * output;
        }

        ii++;
    }

    // Now it's time ti calculate dE/dXn-1

    ii = 0;
    int kk = 0;
    for ( n_iter=m_Neurons.begin(); n_iter<m_Neurons.end(); n_iter++ )
    {
        Neuron& n = *(*n_iter);


        for ( c_iter = n.m_Connections.begin();
              c_iter < n.m_Connections.end(); c_iter++ )
        {
            kk = (*c_iter)->NeuronIndex;
            if ( kk != ULONG_MAX )
            {
                ddXprev[kk] += ddYn[ ii ] *
                       m_Weights[(*c_iter)->WeightIndex ]->value;
            }

        }

        ii++;  // ii tracks the neuron iterator
    }


    double oldValue = 0;
    double newValue = 0;
    for (int jj=0; jj < m_Weights.size(); ++jj )
    {
        oldValue = m_Weights[ jj ]->value;
        newValue = oldValue - etaLearningRate * ddWn[ jj ];
        m_Weights[jj]->value = newValue;
    }
}


// Network BackPropagation

void NeuralNetwork::Backpropagate(double *actualOutput,
     double *desiredOutput)
{
    // Backpropagates through the neural net

    // Proceed from the last layer to the first, iteratively

    VectorLayers::iterator l_iter = m_Layers.end() - 1;

    vector< double > ddXlast( (*l_iter)->m_Neurons.size() );
    vector<vector< double > > differentials;

    int iSize = m_Layers.size();

    differentials.resize( iSize );

    int ii;

    for ( ii=0; ii<(*l_iter)->m_Neurons.size(); ++ii )
    {
        ddXlast[ ii ] =
            actualOutput[ ii ] - desiredOutput[ ii ];
    }

    // We can calculate dE/dXlast straight away after FeedForward, then we need reserve space for other layers


    differentials[iSize-1] = ddXlast;  // last one


    for ( ii = 0; ii < iSize-1; ++ii )
    {
        differentials[ ii ].resize(m_Layers[ii]->m_Neurons.size(), 0.0 );
    }


    ii = iSize - 1;
    for ( l_iter; l_iter > m_Layers.begin(); l_iter--)
    {
        (*l_iter)->Backpropagate( differentials[ ii ],
              differentials[ii - 1], m_etaLearningRate );
        --ii;
    }

    differentials.clear();
}

double Err(double *actualOutput, double *desiredOutput)
{
	double tmp = 0;
	for (int i = 0; i < 10; i++) tmp += (actualOutput[i] - desiredOutput[i]) * (actualOutput[i] - desiredOutput[i]);
	return tmp / 2;
}


int NN_Train(NeuralNetwork *nn, double eps, int iCountMax, TrainingSet *TS, int n, int *npl)
{
    TrainingSet::iterator ts_iter = TS->begin();
    nn->Init(n, npl);
    double *outTmp;
    nn->FeedForward((*ts_iter)->inputVector, 589, outTmp, 10);
    int k = 0;
    while ((k < iCountMax) && (Err((*ts_iter)->desiredOutput, outTmp) > eps))
    {

    }
	return 0;
}

